public struct TicTacToeRules: Rules {
    private let players: [Player]
    public let columns: Int, rows: Int, minAligned: Int
    
    public var type: RulesTypes = .TicTacToe
    
    public init?(players: [Player]) {
        self.init(columns: 3, rows: 3, players: players)
    }
    
    public init?(columns: Int, rows: Int, minAligned: Int = 3, players: [Player]) {
        guard players.count >= 2 else { return nil }
        
        self.columns = columns
        self.rows = rows
        self.minAligned = minAligned
        self.players = players
    }
    
    public func createBoard() -> Board {
        Board(columns: self.columns, rows: self.rows)!
    }
    
    public func isValid(board: Board) -> Bool {
        abs(board.countPieces(filter: { p in p.type == .A }) -
            board.countPieces(filter: { p in p.type == .B })) <= 1
    }
    
    public func isValid(board: Board, move: Move) -> Bool {
        guard self.isValid(board: board) else { return false }
        
        if case .InsertAt(let coord) = move.action, board.isInBounds(coord) {
            return true
        }
        
        return false
    }
    
    public func validMoves(board: Board) -> [Move] {
        return self.players.flatMap { player in
            self.validMoves(board: board, for_player: player).map { Move(player: player, action: $0) }
        }
    }
    
    private static func nextPlayer(board: Board) -> PieceType {
        if board.countPieces(filter: { p in p.type == .A }) > board.countPieces(filter: { p in p.type == .B }) {
            .B
        } else {
            .A
        }
    }
    
    public func validMoves(board: Board, for_player player: Player) -> [Move.Action] {
        var moves: [Move.Action] = []
        
        for col in 0..<board.columns {
            for row in 0..<board.rows {
                if board[col, row] == nil {
                    moves.append(.InsertAt(where: Coords(col, row)))
                }
            }
        }
        
        return moves
    }
    
    public func gameState(board: Board, last_turn: Player?) -> GameState {
        FourInARowRules.rowOfLengthWin(board: board, last_turn: last_turn, minimum_aligned: self.minAligned, players: self.players)
    }
    
    /* public mutating func onMoveDone(move: Move, board: Board) {
        self.history.append(move)
        
        guard case .InsertAt(let coords) = move.action else {
            fatalError("Illegal move \(move.action)")
        }
        
        if FourInARowRules.countMaxRow(center: coords, board: board) >= self.minAligned {
            self.state = .Finished(winner: move.player)
        } else if board.countPieces() == board.columns * board.rows {
            self.state = .Finished(winner: nil)
        } else {
            let next: PieceType = switch move.player {
            case .A: .B
            case .B: .A
            }
            self.state = .Playing(turn: next)
        }
    } */
}
